SERVICE_SUMMARY_MOCK = {
    "documentId": "683",
    "services": [
        {
            "c": "101",
            "dvbtriplet": "2.2045.6301",
            "schedule": True,
            "servicetype": "DSAT",
            "sf": "sd",
            "sg": 12,
            "sid": "2002",
            "sk": 2002,
            "t": "BBC One Lon",
            "xsg": 3
        },
        {
            "c": "0214",
            "dvbtriplet": "2.2095.55128",
            "servicetype": "DSAT",
            "sf": "au",
            "sg": 20,
            "sid": "2862",
            "sk": 2862,
            "t": "UCB Ireland",
            "xsg": 4
        }
    ],
    "version": 3
}

SERVICE_DETAIL_1 = {
    "details": {
        "dvbtriplet": "2.2045.6301",
        "isbroadcasting": True,
        "upgradeMessage": "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    },
    "sid": "2002",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3A7D2/profileAnExample.ts"
        }
    ]
}

SERVICE_DETAIL_2 = {
    "details": {
        "dvbtriplet": "2.2095.55128",
        "isbroadcasting": True,
        "upgradeMessage": "A channel for God's Peace."
    },
    "sid": "2862",
    "streamingprofiles": [
        {
            "name": "AnExample",
            "suri": "http://10.0.1.6:4730/trans_caption/CHAN%3Alocator%3A5%3A3%3AB2E/profileAnExample.ttml",
            "uri": "http://10.0.1.6:4730/trans/CHAN%3Alocator%3A5%3A3%3AB2E/profileAnExample.ts"
        }
    ]
}

SERVICE_MOCK = [SERVICE_SUMMARY_MOCK,SERVICE_DETAIL_1,SERVICE_DETAIL_2]

REMOTE_TCP_MOCK = [  # set up the data to be returned on each successive call of socket.recv()
    b'SKY 000.001\n',
    b'\x01\x01',
    b'\x00\x00\x00\x00',
    b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00',
]



import asyncio

import pytest

from pyskyq.channel import Channel

from .mock_constants import SERVICE_DETAIL_1, SERVICE_SUMMARY_MOCK


def test_channel():


    chan = Channel(SERVICE_SUMMARY_MOCK['services'][0])

    assert isinstance(chan, Channel)
    assert chan.c == "101"
    assert chan.t == "BBC One Lon"
    assert chan.name == "BBC One Lon"

    with pytest.raises(AttributeError, match="Can't modify sid"):
        chan.sid = 232

    with pytest.raises(KeyError, match="blah"):
        chan.blah

    with pytest.raises(KeyError, match="isbroadcasting"):
        chan.isbroadcasting


    chan.add_detail_data(SERVICE_DETAIL_1)

    assert chan.isbroadcasting

    assert chan.upgradeMessage == "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."
    assert chan.desc == "BBC ONE for Greater London and the surrounding area. Find out more about this and the other BBC English regions at www.bbc.co.uk/england."



import pytest
import asyncio

from asynctest import CoroutineMock, MagicMock

from pyskyq import EPG

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import REMOTE_TCP_MOCK, SERVICE_MOCK


def test_EPG(mocker):

    a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.json = CoroutineMock(side_effect=SERVICE_MOCK)

    asyncio.set_event_loop(asyncio.new_event_loop())

    epg = EPG('host')

    assert isinstance(epg, EPG)
    assert len(epg._channels) == 2
    assert epg.get_channel(2002).c == "101"
    assert epg.get_channel(2002).t == "BBC One Lon"
    assert epg.get_channel(2002).name == "BBC One Lon"
    assert epg.get_channel('2002').c == "101"
    assert epg.get_channel('2002').t == "BBC One Lon"
    assert epg.get_channel('2002').name == "BBC One Lon"

    with pytest.raises(AttributeError, match='Channel not found. sid = 1234567.'):
        epg.get_channel(1234567)



"""Import and merge all modules in this package."""
from .skyq import SkyQ
from .constants import RCMD, REMOTE_COMMANDS
from ._version import __version__
from .skyremote import SkyRemote
from .epg import EPG



"""This module implements the EPG class"""

import asyncio
import logging
from typing import Any, List, Optional

from aiohttp import ClientSession, ClientTimeout  # type: ignore

from .channel import Channel
from .constants import (REST_PORT, REST_SERVICE_DETAIL_URL_PREFIX,
                        REST_SERVICES_URL)


class EPG:
    """ The top-level class for all EPG data and functions.

    EPG implements access to any data and operations related to the SkyQ box's
    Electronic Programme Guide. Currently, this means that both summary and detail
    channel information is aggregated into a :class:`pyskyq.channel.Channel` object.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        port (int): Port number to use to connect to the REST HTTP server.
            Defaults to the standard port used by SkyQ boxes which is 9006.
        logger (logging.Logger): Standard Python logger object which if not passed will
            instantiate a local logger.

    """

    def __init__(self,
                 host: str,
                 port: int = REST_PORT,
                 logger: Optional[logging.Logger] = None,
                 ) -> None:
        """Initialise Sky EPG Object.

        This method instantiates the EPG object and populates it with:

        - Channel data

        Args:
            host (str): String with resolvable hostname or IPv4 address to SkyQ box.
            port (int, optional): Port number to use to connect to the Remote REST API.
                Defaults to the standard port used by SkyQ boxes which is 9006.
            logger (logging.Logger, optional): Standard Python logger object which if not
                passed will instantiate a local logger.
        Returns:
            None
        """

        self.host: str = host
        self.port: int = port
        self.logger: logging.Logger = logging.getLogger(__name__) if not logger else logger
        self._channels: list = []
        self.logger.debug(f"Initialised EPG object object with host={host}, port={port}")

        self.load_channels()


        # get list of channels from channel URL
        # get detail's metadata for channels URL
        # create channel's list (in panda's)

    @staticmethod
    async def _fetch(session: ClientSession,
                     url: str
                     ) -> Any:
        """Fetch data from URL asynchronously.

        This helper method fetches data from a URL asynchronously. It is used to fetch EPG
        data from the SkyQ box, including calling the detail endpoint for each channel.

        Args:
            session (aiohttp.ClientSession): Session to use when fetching the data.
            url (str): URL to fetch.

        Returns:
            any: The body of data returned.

        """

        async with session.get(url) as response:
            #TODO add validation etc.
            return await response.json()

    async def _fetch_all_chan_details(self,
                                      session: ClientSession,
                                      sid_list: List[int]
                                      ) -> List:
        """Fetch channel detail data from SkyQ box asynchronously.

        This method fetches the channel list from ``/as/services/detail/<sid>`` endpoint.

        Args:
            session (aiohttp.ClientSession): Session to use when fetching the data.
            sid_list (list): List of Channel SID's to fetch.

        Returns:
            List: List of JSON documents for each channel detail fetched.

        """

        urls = [f'http://{self.host}:{self.port}{REST_SERVICE_DETAIL_URL_PREFIX}{sid}'
                for sid in sid_list]
        results = await asyncio.gather(*[asyncio.create_task(self._fetch(session, url))
                                         for url in urls])
        return results


    async def _load_channel_list(self) -> None:
        """Load channel data into channel property.

        This method fetches the channel list from ``/as/services`` endpoint and load
        :attr:`~pyskyq.epg.EPG._channels`

        Returns:
            None
        """
        self.logger.debug('Fetching channel list')
        url = f'http://{self.host}:{self.port}{REST_SERVICES_URL}'

        timeout = ClientTimeout(total=60)
        async with ClientSession(timeout=timeout) as session:
            chan_payload = await self._fetch(session, url)

        for channel in chan_payload['services']:
            self._channels.append(Channel(channel))


    async def _load_channel_details(self) -> None:
        """Load channel details onto channel properties.

        This method is a wrapper which calls :meth:`~pyskyq.epg.EPG._fetch_all_chan_details`
        to get the details about each channel from it's detail endpoint
        ``/as/services/details/<sid>`` and then adds it to the list data
        on :attr:`~pyskyq.epg.EPG.channels`.

        Returns:
            None
        """
        sid_list = [chan.sid for chan in self._channels]
        timeout = ClientTimeout(total=60)
        async with ClientSession(timeout=timeout) as session:
            channels = await self._fetch_all_chan_details(session, sid_list)
            for channel, sid in zip(channels, sid_list):
                self.get_channel(sid).add_detail_data(channel)


    def load_channels(self) -> None:
        """Load all channel data.

        This is the high-level method that fully loads all the channel detail.

        Returns:
            None
        """

        loop = asyncio.get_event_loop()

        loop.run_until_complete(self._load_channel_list())
        loop.run_until_complete(self._load_channel_details())

        loop.close()

    def get_channel(self,
                    sid: Any
                    ) -> Channel:
        """Get channel data.

        This method returns a :class:`pyskyq.channel.Channel` object when
        passed in a channel ``sid``.

        Args:
            sid: The sid (service id) of the channel
        Returns:
            :class:`pyskyq.channel.Channel`: The channel if found.

        Raises:
            AttributeError: If the channel is not found.
        """
        sid = str(sid)
        for chan in self._channels:
            if chan.sid == sid:
                return chan
        raise AttributeError(f"Channel not found. sid = {sid}.")



import logging
import asyncio
import pytest
from asynctest import CoroutineMock

from pyskyq.constants import REMOTE_PORT, REST_PORT
from pyskyq.skyq import SkyQ

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import SERVICE_MOCK

@pytest.mark.parametrize("host,remote_port,rest_port,logger", [
    ("hostname", REMOTE_PORT, REST_PORT, None),
    ("hostname", 2342, 3243, logging.getLogger()),
])
def test_SkyQ_init(mocker, host, remote_port, rest_port, logger):

    a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.json = CoroutineMock(side_effect=SERVICE_MOCK)

    asyncio.set_event_loop(asyncio.new_event_loop())

    skybox = SkyQ(host, remote_port=remote_port, rest_port=rest_port, logger=logger)

    assert skybox.host == host
    assert skybox.remote_port == remote_port
    assert isinstance(skybox.logger, logging.Logger)




"""This module houses the main SKY Q class"""

import logging
from typing import Optional

from .constants import REMOTE_PORT, REST_PORT
from .skyremote import SkyRemote
from .epg import EPG

# pylint: disable=too-few-public-methods
class SkyQ:
    """Main Sky Q class definition.

    This is the main SkyQ Class definition which provides all available access
    to the SkyQ box that is currently implemented.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        remote_port (int): Port number to use to connect to the Remote TCP socket.
            Defaults to the standard port used by SkyQ boxes which is 49160.
        rest_port (int, optional): Port number to use to connect to the Remote REST API.
            Defaults to the standard port used by SkyQ boxes which is 9006.
        logger (logging.Logger): Standard Python logger object which if not passed will
            instantiate a local logger.
        remote (SkyRemote): An instance of the SkyRemote class which can be used to
            call the lower-level remote API which essentially emulated button-presses on
            the SkyQ Remote.

    Todos:
        * Add interface to read EPG
        * Add higher level channel selector interface to pick channels by name, rather than number.
        * Add misc higher level interfaces to give "now and next" type responses.
    """


    def __init__(self,
                 host: str,
                 *,
                 remote_port: int = REMOTE_PORT,
                 rest_port: int = REST_PORT,
                 logger: Optional[logging.Logger] = None,
                 ) -> None:

        """Initialise SkyQ object.

        This method instantiates the SkyQ object, which will also instantiate a SkyRemote object.

        Args:
            host (str): String with resolvable hostname or IPv4 address to SkyQ box.
            remote_port (int, optional): Port number to use to connect to the Remote TCP socket.
                Defaults to the standard port used by SkyQ boxes which is 49160.
            rest_port (int, optional): Port number to use to connect to the Remote REST API.
                Defaults to the standard port used by SkyQ boxes which is 9006.
            logger (logging.Logger, optional): Standard Python logger object which if not passed
                will instantiate a local logger.
        Returns:
            None

        """
        self.host = host
        self.remote_port = remote_port
        self.rest_port = rest_port
        self.logger = logging.getLogger(__name__) if not logger else logger
        self.remote = SkyRemote(self.host, self.remote_port, self.logger)
        self.epg = EPG(self.host, self.rest_port, self.logger)

        self.logger.debug(f"Initialised SkyQ object with host={host}.")



import asyncio
import socket
import sys

import aiohttp
import pytest
from asynctest import CoroutineMock, MagicMock

from pyskyq import RCMD
from pyskyq.cli import run

from .asynccontextmanagermock import AsyncContextManagerMock
from .mock_constants import SERVICE_MOCK, REMOTE_TCP_MOCK


def test_cli(mocker):

    m = mocker.patch('socket.socket')
    m.return_value.recv.side_effect = REMOTE_TCP_MOCK

    a = mocker.patch('aiohttp.ClientSession.get', new_callable=AsyncContextManagerMock)
    a.return_value.__aenter__.return_value.json = CoroutineMock(side_effect=SERVICE_MOCK)

    mocker.patch.object(sys, 'argv', ['pyskyq', 'play'])

    run()

    m.assert_called_with(socket.AF_INET, socket.SOCK_STREAM)

    assert m.return_value.recv.call_count == 4
    m.return_value.send.assert_any_call(b'SKY 000.001\n')
    m.return_value.send.assert_any_call(b'\x01')
    m.return_value.send.assert_any_call(b'\x00')
    m.return_value.send.assert_any_call(bytearray(b'\x04\x00\x00\x00\x00\x00\xe4\x00'))

    # TODO test with --verbose



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Cli wrapper for the skyq API.

This module implements a cli client which can be used to interact with the SkyQ Box.

Example:
    You can invoke the cli client like this:

        ``$ pyskyq pause``

        ``$ pyskyq play``
Note:
    The main purpose of this package is to provide the API, so the CLI implementation is really
    more just for my own testing. A much better cli client should probably be built using this
    libary, but that is not the scope of this project.

"""

import argparse
import sys
import logging
from typing import List

from pyskyq import __version__, SkyQ, REMOTE_COMMANDS

__author__ = "Bradley Wood"
__copyright__ = "Bradley Wood"
__license__ = "mit"

LOGGER = logging.getLogger(__name__)


def parse_args(args: List[str]) -> argparse.Namespace:
    """Parse command line parameters

    Args:
      args ([str]): command line parameters as list of strings

    Returns:
      :obj:`argparse.Namespace`: command line parameters namespace.
    """
    parser = argparse.ArgumentParser(
        description="Sends a command to a SkyQ box")
    parser.add_argument(
        '--version',
        action='version',
        version='pyskyq {ver}'.format(ver=__version__))
    parser.add_argument(
        dest="cmd",
        help="command to send",
        type=str,
        metavar="CMD")
    parser.add_argument(
        '-v',
        '--verbose',
        dest="loglevel",
        help="set loglevel to INFO",
        action='store_const',
        const=logging.INFO)
    parser.add_argument(
        '-vv',
        '--very-verbose',
        dest="loglevel",
        help="set loglevel to DEBUG",
        action='store_const',
        const=logging.DEBUG)
    return parser.parse_args(args)


def setup_logging(loglevel: int):
    """Setup basic logging

    Args:
      loglevel (int): minimum loglevel for emitting messages.
    """
    logformat = "[%(asctime)s] %(levelname)s:%(name)s:%(message)s"
    logging.basicConfig(level=loglevel, stream=sys.stdout,
                        format=logformat, datefmt="%Y-%m-%d %H:%M:%S")


def main(args: List[str]):
    """Main entry point allowing external calls

    Args:
      args ([str]): command line parameter list
    """
    pargs = parse_args(args)
    setup_logging(pargs.loglevel)
    LOGGER.debug("Starting SkyQ...")
    skyq = SkyQ('skyq')
    skyq.remote.send_command(REMOTE_COMMANDS[pargs.cmd])
    print(skyq.epg.get_channel(2002).desc)
    LOGGER.info("Script ends here")


def run():
    """Entry point for console_scripts.
    """
    main(sys.argv[1:])


if __name__ == "__main__":
    run()



from asynctest import MagicMock

class AsyncContextManagerMock(MagicMock):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        for key in ('aenter_return', 'aexit_return'):
            setattr(self, key,  kwargs[key] if key in kwargs else MagicMock())

    async def __aenter__(self):
        return self.aenter_return

    async def __aexit__(self, *args):
        return self.aexit_return



"""This module implements the SkyRemote class"""

import logging
import socket
import math
from typing import Optional

from .constants import REMOTE_PORT

# pylint: disable=too-few-public-methods
class SkyRemote:
    """Main SkyRemote implementation.

    SkyRemote is a lower level interface to the SkyQ box that seeks to emulate most of the the
    button presses available on the SkyQ remote. There are some buttons which are not (currently)
    supported, however.

    Attributes:
        host (str): Hostname or IPv4 address of SkyQ Box.
        port (int): Port number to use to connect to the Remote TCP socket.
            Defaults to the standard port used by SkyQ boxes which is 49160.
        logger (logging.Logger): Standard Python logger object which if not passed will
            instantiate a local logger.
    Todos:
        * Research some of the missing button codes to see if it's possible to also include them.
    """

    def __init__(self,
                 host: str,
                 port: int = REMOTE_PORT,
                 logger: Optional[logging.Logger] = None,
                 ) -> None:
        """Initialise Sky Remote Object

        This method instantiates the SkyRemote object.

        Args:
            host (str): String with resolvable hostname or IPv4 address to SkyQ box.
            port (int, optional): Port number to use to connect to the Remote TCP socket.
                Defaults to the standard port used by SkyQ boxes which is 49160.
            logger (logging.Logger, optional): Standard Python logger object which if not
                passed will instantiate a local logger.
        Returns:
            None
        """

        self.host = host
        self.port = port
        self.logger = logging.getLogger(__name__) if not logger else logger
        self.logger.debug(f"Initialised SkyRemote object with host={host}, port={port}")

    def send_command(self, code: int) -> None:
        """Send a command to the Sky Q box using the Remote interface.

        This method sends a button press to the SkyQ box via a TCP socket connection. The
        encoding and handshake required was extracted by looking at the following sources:

        - https://github.com/dalhundal/sky-remote
        - https://gladdy.uk/blog/2017/03/13/skyq-upnp-rest-and-websocket-api-interfaces/

        Args:
            code (int): A code passed which represents the button to press. See
                :py:mod:`pyskyq.constants` for a human-friendly list of the currently
                supported buttons and their associated codes.

        Returns:
            None

        """


        self.logger.debug(f'Sending command code={code}')
        command_bytes = bytearray([4, 1, 0, 0, 0, 0, math.floor(224 + (code/16)), code % 16])
        self.logger.debug(f'command_bytes={command_bytes}')

        # TODO try/except socket.gaierror
        client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        client.connect((self.host, self.port))

        # Wait until ready for data
        length = 12
        while True:
            data = client.recv(24)
            self.logger.debug(f'Received data={data}')
            if len(data) < 24: # box sends 24 x \x00 to signal end of hand-shake.
                client.send(data[0:length])
                self.logger.debug(f'Sent data={data[0:length]}')
                length = 1
            else:
                break

        client.send(command_bytes)
        self.logger.debug(f'Send command part 1 data={command_bytes}')
        command_bytes[1] = 0
        client.send(command_bytes)
        self.logger.debug(f'Send command part 2 data={command_bytes}')
        client.close()



"""Constants for button codes, ports, and so on"""
from collections import namedtuple
from typing import Dict

REMOTE_LEGACY_PORT: int = 5900
"""int: Legacy port number for older Sky Boxes."""

REMOTE_PORT: int = 49160
"""int: Current port number for newer Sky Boxes."""

REMOTE_COMMANDS: Dict[str, int] = {
    'power': 0,
    'select': 1,
    'backup': 2,
    'dismiss': 2,
    'channelup': 6,
    'channeldown': 7,
    'interactive': 8,
    'sidebar': 8,
    'help': 9,
    'services': 10,
    'search': 10,
    'tvguide': 11,
    'home': 11,
    'i': 14,
    'text': 15,
    'up': 16,
    'down': 17,
    'left': 18,
    'right': 19,
    'red': 32,
    'green': 33,
    'yellow': 34,
    'blue': 35,
    'zero': 48,
    'one': 49,
    'two': 50,
    'three': 51,
    'four': 52,
    'five': 53,
    'six': 54,
    'seven': 55,
    'eight': 56,
    'nine': 57,
    'play': 64,
    'pause': 65,
    'stop': 66,
    'record': 67,
    'fastforward': 69,
    'rewind': 71,
    'boxoffice': 240,
    'sky': 241,
}
"""
Dictionary mapping strings of button names to their associated codes.
"""  # pylint: disable=W0105

# Silence mypy type errors on the below 2 commands.

RCMDTUPLE = namedtuple('RCMDTUPLE', sorted(REMOTE_COMMANDS))  # type: ignore

RCMD = RCMDTUPLE(**REMOTE_COMMANDS)  # type: ignore
"""
Named tuple object with attributes as button names and their values as the associated remote code.
"""  # pylint: disable=W0105

REST_PORT: int = 9006
"""int: Port number for the REST API."""

REST_SERVICES_URL: str = '/as/services'
"""str: REST endpoint for list of services (channels)."""

REST_SERVICE_DETAIL_URL_PREFIX: str = '/as/services/details/'
"""str: REST endpoint for service (channel) details."""

CHANNEL_FIELD_MAP = {
    "number": "c", # "101",
    "quality": "sf", # "sd",
    "id": "sid", # "2002",
    "name": "t",  # "BBC One Lon",
    "desc": "upgradeMessage",
}
"""Dict(str,str): Maps human-friendly names to Sky's terse field names (where understood)."""



"""This module implements the Channel class"""
import logging
from typing import Dict, Optional, Any

from .constants import CHANNEL_FIELD_MAP

class Channel:
    """ This class holds channel data and methods for manipulating the channel.

        As the properties served from the SkyQ REST API are immutable, this class
        presents these as read-only attributes on the object. It will also dynamically
        set attributes based on the payload obtained from the API, which itself could vary
        from one channel to another. For example, channel's which are '+1' all carry the
        ``timeshifted`` property whereas regular channels do not.

        Finally, in order to present a more human-friendly API the
        :const:`~pyskyq.constants.CHANNEL_FIELD_MAP` dictionary is used to provide access to
        properties using more friendly names.

        Note:
            These channel attributes could change at any time with a box upgrade. The API will
            attempt to adapt to ensure the new fields are presented.

        Attributes:
            c (str): Channel number.
            dvbtriplet (str):  DVB Triplet (I have no idea what this is)
            schedule (bool): Is the channel "scheduled" or not?
            servicetype (str): Where is the channel coming from (e.g., ``DSAT``)
            sf (str): Quality of the channel (e.g., ``hd``)
            sg (int): No ideas what this is.
            sid (str): Channel id (aka ``sid``) in `str` form - the **primary key** for the channel.
            sk (int): Channel id (aka ``sid``) in `str` form.
            t (str): Channel name, eg, ``BBC One Lon``
            xsg (int): No idea what this is.
            timeshifted (bool): Is this a ``+1``-style channel?
            upgradeMessage (str): A short description of the channel.
            isbroadcasting (boo): Is the channel broadcasting currently or not?

        Note:
            The below are the human-friendly mappings of some of Sky's terser-named
            attributes.

        Attributes:
            number (str): Human-friendly version of :attr:`c`.
            quality (str): Human-friendly version of :attr:`sf`.
            id (str): Human-friendly version of :attr:`sid`.
            name (str):Human-friendly version of :attr:`t`.
            desc (str): Human-friendly version of :attr:`upgradeMessage`.


    """

    def __init__(self,
                 chan_dict: Dict,
                 logger: Optional[logging.Logger] = None,
                 ) -> None:
        """Initialise Channel Object.

        Args:
            chan_dict (dict): This dictionary is the payload that comes directly from the
                ``as/services/`` endpoint.
            logger (logging.Logger, optional): Standard Python logger object which if not
                passed will instantiate a local logger.
        Returns:
            None

        """
        self._chan_dict = chan_dict
        self.logger = logger
        if self.logger:
            self.logger.debug(f"Channel {self._chan_dict['t']} instantiated.")

    def __getattr__(self, name: str) -> Any:
        if name in self._chan_dict.keys():
            return self._chan_dict[name]
        if name in CHANNEL_FIELD_MAP.keys():
            return self._chan_dict.get(CHANNEL_FIELD_MAP[name], None)
        raise KeyError(name)


    def __setattr__(self, name: str, value: Any) -> None:
        if not name.startswith('_') and name != 'logger':
            raise AttributeError(f"Can't modify {name}")
        else:
            super().__setattr__(name, value)


    def add_detail_data(self, detail_dict: Dict) -> None:
        """ Add additional properties obtained from the detail endpoint to the object.

            Args:
                detail_dict (dict): A detail dict that is passed in directly from the
                    ``as/services/details/<sid>`` endpoint.
            Returns:
                None

        """
        self._chan_dict.update(detail_dict['details'])
